/*
	ProWesS reader - a HTML file browser

	PROforma, PROGS Font & Raster Manager
	ProWesS, PROGS Windowing System

	definition of a HyperText object

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 18 November 1995
	24.11.1995 word wrap added
	25.11.1995 buffer full now handled properly
	19.06.1996 HTImage added
	28.06.1996 display colour re-set after displaying an image
*/

#include "HyperText_h"

/* forward declarations */
static Error LineBreak(HyperText *text);
static Error LinkTerminate(HyperText *text);

static ERROR_BUILD(ErrSize,"File too long to display");
#define INCR_HEIGHT(x) { pt keep=text->height+(x); \
						 if (keep>short2pt(32000)) return ERROR_CODE(ErrSize); \
						 else text->height=keep; }
extern char PaperWhite;

Error HTReInit(Global *global, Gstate gstate, Style *look, HyperText **text, pt columnwidth)
{
	Error err;
	Heap heap;
	HyperText *ht;
	StylePage *style;
	StyleLine *font;
	Chunks cont_chunks=NUL;
	Contents *cont_first=NULL,*cont_curr=NULL;

	/* release old HyperText object if there was one */
	if (*text)
	{
		cont_first=(*text)->cont_first;
		cont_curr=(*text)->cont_curr;
		cont_chunks=(*text)->cont_chunks;

		heap=(*text)->heap;
		*text=NULL;
		MEMHeapZap(heap);
	}

	/* allocate new HyperText object */
	err=MEMHeapInit(HYPERTEXT_BLOCKSIZE, &heap);
	if (err) return err;

	err=MEMHeapAllocate(heap, sizeof(HyperText), &ht);
	if (err) { MEMHeapZap(heap); return err; }

	MEMClear(sizeof(HyperText),ht);
	*text=ht;
	ht->global=global;
	ht->heap=heap;
	ht->gstate=gstate;
	ht->look=look;
	ht->lines.next=ht->lines.prev=&ht->lines;
	ht->cont_chunks=cont_chunks;
	ht->cont_first=cont_first;
	ht->cont_curr=cont_curr;
	ht->cont_insert=cont_curr;
	ht->colour=PaperWhite||global->printing ? pt_hundred : 0;

	/* initialise default stylepage */
	MEMHeapAllocate(heap,sizeof(StylePage),&style);
	MEMClear(sizeof(StylePage),style);
	style->font=font;
	style->wrap=TRUE;
	style->marginright=columnwidth;
	ht->style=style;

	/* select current font and size */
	catch( HTChangeFont(ht,ROMAN,100) );

	/* create the first TextLine element */
	catch( LineBreak(ht) );

	return err;
}

typedef struct { LineElement *next; Error (*Draw)(); char base[2]; } ElementString;

static Error drawstring(Gstate gstate, ElementString *me)
{   return PFStringShow(gstate,me->base); }

static Error AddString(HyperText *text, char *buffer, Size length, pt width)
{
	Error err=ERR_OK;
	ElementString *base;
	TextLine *lastline;
	pt above, below, diff;

	/* allocate some heap memory for the text */
	catch( MEMHeapAllocate(text->heap, sizeof(*base)+length, &base) );
	MEMCopy(length,buffer,base->base);
	base->base[length]='\0';
	base->next=NULL;
	base->Draw=drawstring;

	/* adjust current position in line */
	text->posinline+=width;

	/* link in */
	lastline=text->lines.prev;
	if (lastline->last)
		lastline->last->next=(LineElement *)base;
	else
		lastline->elements=(LineElement *)base;
	lastline->last=(LineElement *)base;

	/* adjust height of current line */
	above=text->style->baseline;
	below=text->style->height-text->style->baseline;
	/* adjust space above baseline */
	if ( (diff = above - lastline->baseline) > 0 )
	{
		lastline->height+=diff;
		lastline->baseline+=diff;
	}
	/* and below baseline */
	if ( (diff = below - (lastline->height-lastline->baseline)) > 0 )
		lastline->height+=diff;

	return err;
}

static Error BufferFlush(HyperText *text)
{
	Error err=ERR_OK;

	if (text->bufferpos==0) return err;

	if (text->style->wrap)
	{
		char *pos=text->buffer, oldch;
		char *cpos=pos, *oldpos;
		pt width, prevwidth;
		Size prevlen;

		text->buffer[text->bufferpos]='\0';
		prevwidth=0; prevlen=0;

		while (*cpos)
		{
			oldpos=cpos;

			/* scan for end of this word and get width */
			for ( ; *cpos && *cpos!=' ' ; cpos++);
			if (*cpos) cpos++;
			oldch=*cpos;
			*cpos='\0';
			PFStringWidth(text->gstate,oldpos,&width);
			*cpos=oldch;

			if ( text->style->marginright < text->posinline+prevwidth+width )
			{
				/* add all the previous words to current line */
				catch( AddString(text,pos,prevlen,prevwidth) );

				/* newline */
				catch( LineBreak(text) );

				/* prepare for next piece of buffer */
				pos=oldpos;
				prevwidth=0;
				prevlen=0;
			}
			prevwidth+=width;
			prevlen+=cpos-oldpos;
		}
		err=AddString(text,pos,prevlen,prevwidth);
	}
	else
	{
		pt width;

		text->buffer[text->bufferpos]='\0';
		PFStringWidth(text->gstate, text->buffer, &width);
		err=AddString(text, text->buffer, text->bufferpos, width);
	}
	/* buffer is empty now */
	text->bufferpos=0;

	return err;
}

Error HTAddCharacter(HyperText *text, char ch)
{
	Error err=ERR_OK;

	/* if room in buffer -> add extra char, done */
	if (text->bufferpos>BUFFERSIZE)
	{
		char ch;
		int space;

		/* adjust size of buffer not to include anything after last space */
		for (space=BUFFERSIZE ; space>0 && text->buffer[space]!=' ' ; space--);
		if (space==0) return ERR_ISYN;
		text->bufferpos=space+1;
		ch=text->buffer[space+1];

		/* flush buffer */
		catch( BufferFlush(text) );

		/* move remainder to start of buffer */
		text->buffer[space+1]=ch;
		text->bufferpos=BUFFERSIZE-space;
		MEMCopy(text->bufferpos,text->buffer+space+1,text->buffer);
	}

	/* add new character */
	text->buffer[text->bufferpos++]=ch;

	return err;
}

typedef struct { LineElement *next; Error (*Draw)(); char *font; pt size; } ElementFont;

static Error chfont(Gstate gstate, ElementFont *me)
{   PFFontLoad(gstate,me->font); return PFFontScale(gstate,me->size);   }

Error HTChangeFont(HyperText *text, FontStyle sfont, short perc_size)
{
	Error err;
	char init=(text->style->font==NULL);
	StyleLine *st;
	pt size;
	Bbox bbox;
	char *font;

	catch( BufferFlush(text) );

	/* make font and size real + update current values */
	if (sfont==PREVIOUS)
		font=text->style->font->font;
	else
		font=text->look->fonts[(short)(sfont-1)];
	if (!perc_size)
		size=text->style->font->size;
	else
		size=text->look->bodyfontsize*perc_size/100;

	/* check whether a suitable StyleLine already exists */
	/* or allocate a new one */
	for (st=text->fonts ; st && (st->font!=font || st->size!=size) ; st=st->next);
	if (!st)
	{
		catch( MEMHeapAllocate(text->heap, sizeof(StyleLine), &st) );
		st->font=font;
		st->size=size;
		st->next=text->fonts;
		text->fonts=st;
	}
	text->style->font=st;

	/* determine extension of font above and below the baseline */
	PFFontLoad(text->gstate,font);
	PFFontScale(text->gstate,size);
	PFFontBbox(text->gstate,&bbox);
	text->style->height=bbox.ysiz;
	text->style->baseline=bbox.yorg;

	/* set element to change font and size */
	if (!init)
	{
		ElementFont *base;
		TextLine *line=text->lines.prev;

		/* allocate some heap memory for the element */
		catch( MEMHeapAllocate(text->heap, sizeof(*base), &base) );
		base->next=NULL;
		base->Draw=chfont;
		base->font=font;
		base->size=size;

		/* link in */
		if (line->last)
			line->last->next=(LineElement *)base;
		else
		{
			line->elements=(LineElement *)base;
		}
		line->last=(LineElement *)base;
	}

	return ERR_OK;
}

typedef struct { LineElement *next; Error (*Draw)(); pt dist; } ElementMoveX;

static Error movex(Gstate gstate, ElementMoveX *me)
{   return PFMoveR(gstate,me->dist,0);  }

static Error MoveXR(HyperText *text, TextLine *line, pt dist)
{
	Error err;
	ElementMoveX *base;

	if (dist==0) return err;
	text->posinline+=dist;

	/* allocate some heap memory for the element */
	catch( MEMHeapAllocate(text->heap, sizeof(*base), &base) );
	base->next=NULL;
	base->Draw=movex;
	base->dist=dist;

	/* link in */
	if (line->last)
		line->last->next=(LineElement *)base;
	else
	{
		line->elements=(LineElement *)base;
	}
	line->last=(LineElement *)base;
	return err;
}

Error HTChangeMargin(HyperText *text, short perc_left, short perc_right)
{
	Error err;
	pt marginleft, marginright;
	pt left=text->look->bodyfontsize*perc_left/100;
	pt right=text->look->bodyfontsize*perc_right/100;

	catch( BufferFlush(text) );

	/* adjust margins in stylepage */
	marginleft =text->style->marginleft +=left;
	marginright=text->style->marginright-=right;
	text->linkstart=marginleft;

	/* adjust remaining width and position in line */
	if (text->posinline!=marginleft)
	{
		catch( MoveXR(text,text->lines.prev,marginleft-text->posinline) );
		text->posinline=marginleft;
	}
	return ERR_OK;
}

static Error LineBreak(HyperText *text)
{
	Error err;
	TextLine *line;

	catch( LinkTerminate(text) );
	text->width=max(text->width,text->posinline);
	INCR_HEIGHT(text->lines.prev->height);

	/* create a new TextLine element */
	catch( MEMHeapAllocate(text->heap, sizeof(TextLine), &line) );
	MEMClear(sizeof(TextLine),line);
	line->styleline=text->style->font;

	/* link in as last line */
	line->next=&text->lines;
	line->prev=text->lines.prev;
	text->lines.prev->next=line;
	text->lines.prev=line;

	/* set left margin position if necessary */
	text->posinline=0;
	if (text->style->marginleft)
		catch( MoveXR(text,line,text->style->marginleft) );

	return err;
}

Error HTLineBreak(HyperText *text)
{
	Error err;

	catch( BufferFlush(text) );
	err=LineBreak(text);
	return err;
}

Error HTSpaceHorizontal(HyperText *text, short perc_dist)
{
	Error err;

	catch( BufferFlush(text) );
	err=MoveXR(text,text->lines.prev,text->look->bodyfontsize*perc_dist/100);
	return err;
}

Error HTSpaceVertical(HyperText *text, short perc_dist)
{
	Error err;
	pt dist=text->look->bodyfontsize*perc_dist/100;

	catch( HTLineBreak(text) );

	/* increase height of previous line by dist */
	text->lines.prev->prev->height+=dist;
	INCR_HEIGHT(dist);

	return err;
}

Error HTStyleNew(HyperText *text)
{
	Error err;
	StylePage *style;

	/* flush buffers */
	catch( BufferFlush(text) );

	/* allocate a new StylePage element - copy and link in */
	catch( MEMHeapAllocate(text->heap,sizeof(StylePage),&style) );
	MEMCopy(sizeof(StylePage),text->style,style);
	style->next=text->style;
	text->style=style;

	return err;
}

Error HTStyleRestore(HyperText *text)
{
	Error err;
	StylePage *style=text->style;

	/* flush buffers */
	catch( BufferFlush(text) );

	/* unlink current style and release previous stylepage */
	text->style=style->next;
	MEMHeapRelease(text->heap,style);
	err=HTChangeFont(text,PREVIOUS,0);
	return err;
}

Error HTEnd(HyperText *text)
{
	return HTLineBreak(text);
}

Error HTLineWrap(HyperText *text, char status)
{
	text->style->wrap=status;
	return ERR_OK;
}

typedef struct { LineElement *next; Error (*Draw)(); pt dist; pt width; } ElementLine;

static Error liner(Gstate gstate, ElementLine *me)
{   PFLineWidth(gstate, me->width); return PFLineR(gstate,me->dist,0);  }

Error HTLineHorizontal(HyperText *text)
{
	Error err;
	ElementLine *base;
	TextLine *line;

	catch( HTLineBreak(text) );

	/* allocate some heap memory for the element */
	catch( MEMHeapAllocate(text->heap, sizeof(*base), &base) );
	base->next=NULL;
	base->Draw=liner;
	base->dist=text->style->marginright-text->posinline;
	base->width=text->look->line_thickness;
	text->width=max(text->style->marginright,text->width);

	line=text->lines.prev;
	/* link in */
	if (line->last)
		line->last->next=(LineElement *)base;
	else
	{
		line->elements=(LineElement *)base;
	}
	line->last=(LineElement *)base;

	line->height=(2*text->look->line_space+text->look->line_thickness)*text->look->bodyfontsize/100;
	line->baseline=(text->look->line_space+text->look->line_thickness/2)*text->look->bodyfontsize/100;

	err=HTLineBreak(text);
	return err;
}

Error HTLinkStart(HyperText *text, char *filename, char *position)
{
	Error err;

	catch( BufferFlush(text) );
	text->linkstart=text->posinline;
	catch( MEMHeapAllocate(text->heap,sizeof(Link),&text->linkcurr) );
	STRCopySafe(filename,text->linkcurr->filename,IO_MAXFULLNAME);
	STRCopySafe(position,text->linkcurr->position,MAXLINKPOSITION);

	return err;
}

Error HTLinkEnd(HyperText *text)
{
	Error err=ERR_OK;

	if (!text->linkcurr) return err;		/* is no link element */

	catch( BufferFlush(text) );
	catch( LinkTerminate(text) );

	/* no longer inside a link */
	text->linkcurr=NULL;

	return err;
}

typedef struct { LineElement *next; Error (*Draw)(); pt xdist; pt ydist; } ElementUnderLine;

static Error underline(Gstate gstate, ElementUnderLine *me)
{
	PFLineWidth(gstate, 0);
	PFMoveR(gstate,0,me->ydist);
	PFLineR(gstate,-me->xdist,0);
	return PFMoveR(gstate,me->xdist,-me->ydist);
}

static Error LinkTerminate(HyperText *text)
{
	Error err=ERR_OK;
	LinkArea *link;
	ElementUnderLine *base;
	TextLine *line;

	if (!text->linkcurr || text->lines.prev->height==0
						|| text->lines.prev->baseline==0) return err;

	catch( MEMHeapAllocate(text->heap,sizeof(LinkArea),&link) );
	link->next=text->lines.prev->links;
	link->xorg=text->linkstart;
	link->xend=text->posinline;
	link->destination=text->linkcurr;
	text->lines.prev->links=link;
	text->linkstart=text->style->marginleft;

	/* the link has to be underlined to make it visible !! */

	/* allocate some heap memory for the element */
	catch( MEMHeapAllocate(text->heap, sizeof(*base), &base) );
	base->next=NULL;
	base->Draw=underline;
	base->xdist=link->xend-link->xorg;
	base->ydist=(text->style->height-text->style->baseline)/2;

	line=text->lines.prev;
	/* link in */
	if (line->last)
		line->last->next=(LineElement *)base;
	else
	{
		line->elements=(LineElement *)base;
	}
	line->last=(LineElement *)base;

	return err;
}

Error HTLinkMarker(HyperText *text, char *position)
{
	Error err=ERR_OK;
	Marker *marker;

	catch( MEMHeapAllocate(text->heap, sizeof(Marker), &marker) );

	marker->next=text->lines.prev->markers;
	STRCopySafe(position,marker->name,MAXLINKPOSITION);
	text->lines.prev->markers=marker;

	return err;
}

Error HTContentsAdd(HyperText *text, char *filename)
{
	Channel file;
	Error err;
	char realfilename[IO_MAXFULLNAME];
	char realdirectory[IO_MAXDIRECTORY];
	Contents *cont;

	/* determine correct filename and directory for file */
	err=IOOpenPath(filename,OPEN_READ,text->directory,&file);
	if (err) return ERR_OK;
	IOName(file,realfilename,realdirectory);
	IOClose(file);

	/* see whether it already exist */
	for (cont=text->cont_first ;
		 cont && !(STRSameCI(cont->filename,realfilename) &&
				   STRSameCI(cont->directory,realdirectory) ) ;
		 cont=cont->next);
	if (cont) return ERR_OK;

	/* add to contents list */
	if (!text->cont_chunks)
	{
		err=MEMUserInit(sizeof(Contents),10,&text->cont_chunks);
		if (err) return ERR_OK;
	}
	err=MEMUserAllocate(text->cont_chunks,&cont);
	if (err) return ERR_OK;
	STRCopy(realfilename, cont->filename);
	STRCopy(realdirectory,cont->directory);

	/* link in */
	if (text->cont_insert)
	{
		cont->next=text->cont_insert->next;
		text->cont_insert->next=cont;
		text->cont_insert=cont;
	}
	else
	{
		cont->next=NULL;
		text->cont_first=cont;
		text->cont_insert=cont;
	}

	return ERR_OK;
}

Error HTContentsDelete(HyperText *text)
{
	Contents *c=text->cont_first, *c1;

	text->cont_first=text->cont_curr=text->cont_insert=NULL;
	while(c)
	{
		c1=c;
		c=c->next;
		MEMUserRelease(text->cont_chunks,c1);
	}
	return ERR_OK;
}

typedef struct {
	LineElement *next;
	Error (*Draw)();
	pt xsize, ysize;
	int id;
	pt colour;
	FileInfo fi;
	char base[2];
} ElementImage;

Error image(Gstate gstate, ElementImage *el)
{
	Error err;

	err=PFPictureDisplay(gstate,el->id,(char *)&el->fi,el->xsize,el->ysize);
	PFColourGray(gstate, el->colour);
	return err;
}

Error HTImage(HyperText *text, char *filename, pt xsize, pt ysize)
{
	Error err;
	ElementImage *base;
	TextLine *line;
	Channel file;
	Size size;
	int id, recognized=FALSE;

	catch( HTLineBreak(text) );

	err=IOOpenPath(filename,OPEN_READ,text->global->directory,&file);
	if (err) return ERR_OK;
	IOLength(file,&size);

	/* allocate some heap memory for the element */
	if (err=MEMHeapAllocate(text->heap, sizeof(*base)+size, &base))
		{ IOClose(file); return err; }
	if (err=IOFileInfoGet(file,&base->fi))
		{ IOClose(file); return err; }
	if (err=IOFileLoad(file,size,base->base))
		{ IOClose(file); return err; }
	IOClose(file);

	/* figure out the picture driver id */
	PFPictureCount(text->gstate,NULL);
	while (!recognized && !PFPictureNext(text->gstate,&id,NULL))
	{
		if (PFPictureRecognize(text->gstate,id,(char *)&base->fi)==ERR_OK)
			recognized=TRUE;
	}
	if (!recognized)
		return MEMHeapRelease(text->heap,base);
	else
		base->id=id;

	if (xsize==0 || ysize==0)
	{
		int xratio, yratio;
		PFPictureRatio(text->gstate,id,(char *)&base->fi,&xratio,&yratio);
		if (xsize==0)
		{
			if (ysize==0)
			{
				xsize=text->global->canvaswidth/2;
				ysize=(pt)((double)xsize/xratio*yratio);
			}
			else
				xsize=(pt)((double)ysize/yratio*xratio);
		}
		else
			ysize=(pt)((double)xsize/xratio*yratio);
	}

	base->next=NULL;
	base->Draw=image;
	base->xsize=xsize;
	base->ysize=ysize;
	base->colour=text->colour;
	text->width=xsize;

	line=text->lines.prev;
	/* link in */
	if (line->last)
		line->last->next=(LineElement *)base;
	else
	{
		line->elements=(LineElement *)base;
	}
	line->last=(LineElement *)base;

	line->height=ysize;
	line->baseline=0;

	err=HTLineBreak(text);
	return err;
}
